def int_to_bytes(i: int) -> bytes:
    """
    Wrapper around int.to_bytes. Set some defaults parameters

    """
    return i.to_bytes(i.bit_length() // 8 + 1, 'little', signed=True)


def flat_list(tree: list) -> list:
    """
    Recursively unnest a nested list
    Args:
        tree: nested lists of unlimited depth

    Returns:
        l: flat list

    """
    l = []
    for node in tree:
        if isinstance(node, list):
            l += flat_list(node)
        else:
            l.append(node)
    return l



# bokeh serve --show taskviewer_app.py

from app.flow.commons import params
from pycarol.luigi_extension import TaskViewer
from app.flow.output import RunMe
from bokeh.plotting import curdoc

TV = TaskViewer(RunMe(**params))

TV.plot_task(curdoc(), web_app=True)



from .task import(
    Task,
    WrapperTask,
    inherit_list,
    inherit_dict
)

from .targets import (
    CDSTarget,
    PickleTarget,
    KerasTarget,
    DummyTarget,
    JsonTarget,
    FeatherTarget,
    PytorchTarget,
    PickleTarget,
    PytorchTarget,
    KerasTarget,
    LocalTarget,
    DummyTarget,
)
from pycarol.luigi_extension.targets.deprecated_targets import PicklePyCarolTarget, PytorchPyCarolTarget, \
    KerasPyCarolTarget, PickleLocalTarget, KerasLocalTarget, PytorchLocalTarget, JsonLocalTarget, FeatherLocalTarget


from pycarol.luigi_extension.taskviewer.visualization import (
    Visualization
)



import numpy as np
from pycarol.luigi_extension.task import Task

from typing import Tuple

def luigi_get_sons(task) -> list:
    """
    Returns a list of required tasks.
    Args:
        task: luigi Task

    Returns:
        l: list of luigi Task

    """
    return task.requires()

def get_dag_from_task(top_nodes: list, get_sons: 'function' = None) -> dict:
    """
    Extract a Direct Acyclic Graph structure of a pipeline using 
    get_sons_method to fetch sons nodes
    
    Args:
        task: list of top tasks/nodes
        get_sons: method to extract sons node (required tasks) of a 
        given node

    Returns:
        dag: dictionary encoding a DAG data structure.
 
    """
    assert isinstance(top_nodes,list)
    dag = {}

    def _traverse_tree(task_list):
        # breadth first search
        nonlocal dag, get_sons

        # add new nodes
        for t in task_list:
            if t not in dag:
                dag[t] = get_sons(t)

        # get all nodes of this level
        sons_list = []
        for k,v in dag.items():
            for vi in v:
                if vi not in dag:
                    sons_list.append(vi)

        # recursion level wise
        if sons_list:
            _traverse_tree(sons_list)

    _traverse_tree(top_nodes)
    return dag





from pytest import mark
from pycarol.luigi_extension.task import Task,inherit_list
from .taskviewer import *

class T1(Task):
    pass

@inherit_list(T1)
class T2(Task):
    pass

@inherit_list(T1,T2)
class T3(Task):
    pass

@inherit_list(T2,T3)
class T4(Task):
    def easy_run(self, inputs):
        raise Exception

@inherit_list(T1,T3)
class T5(Task):
    pass

params = {}


def test_task_run():
    T4(**params).buildme()

def test_get_dag_from_task():
    d = get_dag_from_task([T5(**params)],luigi_get_sons)
    print(d)
